# -*- coding: utf-8 -*-
import six
from importlib import import_module


def import_by_path(dotted_path):
    """Import a dotted module path and return the attribute/class designated
    by the last name in the path. Raise ImportError if the import failed.

    Adapted from Django 1.7
    """

    try:
        if not dotted_path.count('.'):
            dotted_path = '.'.join(['__main__', dotted_path])

        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:
        msg = '"{}" doesn\'t look like a module path'.format(dotted_path)
        raise ImportError(msg)

    try:
        module = import_module(module_path)
    except ImportError as ex:
        raise ImportError('Failed to import "{}" - {}'.format(dotted_path, ex))

    try:
        return getattr(module, class_name)
    except AttributeError:
        msg = 'Module "{}" does not define a "{}" attribute/class'.format(
            dotted_path, class_name)
        raise ImportError(msg)


def get_function_path(function):
    """Get received function path (as string), to import function later
    with `import_string`.
    """
    if isinstance(function, six.string_types):
        return function

    func_path = []

    module = getattr(function, '__module__', '__main__')
    if module:
        func_path.append(module)

    func_path.append(function.__name__)
    return '.'.join(func_path)


def django_setup(settings_module=None):
    """Initialize Django if required, must be run before performing
    any task on spooler or mule.
    """
    from django.conf import settings, ENVIRONMENT_VARIABLE

    if settings.configured:
        return

    if settings_module:
        import os
        os.environ[ENVIRONMENT_VARIABLE] = settings_module

    try:
        # django > 1.7
        from django import setup
    except ImportError:
        # django < 1.7
        def setup():
            settings._setup()

    setup()


class ProxyDict(dict):

    def __init__(self, dict_instance, key):
        super(ProxyDict, self).__init__()

        self.key = key

        if self.key in dict_instance:
            self.update(dict_instance[self.key])

        dict_instance[self.key] = self



# -*- coding: utf-8 -*-
from datetime import timedelta
from unittest import TestCase

import os
import mock
import six
from uwsgi_tasks.utils import import_by_path, get_function_path
from uwsgi_tasks.tasks import (
    RuntimeTask, manage_mule_request, manage_spool_request, TimerTask,
    get_current_task, SpoolerTask
)
from uwsgi_tasks import task, TaskExecutor, RetryTaskException, SPOOL_OK


def local_function(value):
    return value ** 2


class UtilsTest(TestCase):
    builtin_module = '__builtin__' if six.PY2 else 'builtins'

    def test_import_by_path(self):
        with self.assertRaises(ImportError):
            import_by_path('len')

        with self.assertRaises(ImportError):
            import_by_path('uwsgi_tasks.tests.unknown_function')

        main = __import__('__main__')
        main.local_function = local_function

        self.assertTrue(local_function is import_by_path('local_function'))

        self.assertTrue(
            local_function is import_by_path('uwsgi_tasks.tests.local_function')
        )
        self.assertTrue(len is import_by_path('{}.len'.format(
            self.builtin_module)))

    def test_get_function_path(self):
        self.assertEqual(
            get_function_path(local_function),
            'uwsgi_tasks.tests.local_function'
        )

        self.assertEqual(
            get_function_path(len),
            '{}.len'.format(self.builtin_module)
        )

        def nested():
            pass

        self.assertEqual(get_function_path(nested), 'uwsgi_tasks.tests.nested')


storage = mock.MagicMock()


@task(executor=TaskExecutor.MULE)
def mule_task(a, b, c='3', d='4'):
    storage(a, b, c, d)


@task(executor=TaskExecutor.SPOOLER, priority='10', at=timedelta(seconds=10))
def spooler_task(a, b=None):
    storage(a, b)


@task(executor=TaskExecutor.SPOOLER, retry_count=2)
def spooler_retry_task(g, h):
    storage(g, h)
    raise RetryTaskException(timeout=10)


@task(executor=TaskExecutor.SPOOLER, retry_count=2,
      retry_timeout=timedelta(seconds=20))
def spooler_and_task_introspection(a, b):
    current_task = get_current_task()
    assert isinstance(current_task, SpoolerTask)
    assert current_task.executor == TaskExecutor.SPOOLER

    if len(current_task.buffer):
        assert current_task.retry_count == 1
        assert current_task.retry_timeout == 50
        assert current_task.at == timedelta(seconds=50)
        assert len(current_task.buffer) == 1
        assert current_task.buffer[b'message'] == u'Hello, World!'
    else:
        assert current_task.executor == TaskExecutor.SPOOLER
        assert current_task.retry_count == 2
        assert current_task.retry_timeout == 20

    storage(a, b)
    current_task.buffer[b'message'] = u'Hello, World!'
    raise RetryTaskException(timeout=50)


SPOOLER_DIR = os.path.expanduser('~')


@task(executor=TaskExecutor.SPOOLER)
def spooler_task_with_cwd_changed():
    storage()

    current_dir = os.getcwd()
    assert current_dir != SPOOLER_DIR


@task(executor=TaskExecutor.SPOOLER, working_dir=None)
def spooler_task_with_cwd_not_changed():
    storage()

    current_dir = os.getcwd()
    assert current_dir == SPOOLER_DIR


def timer_task(signum):
    storage(signum)


class TaskTest(TestCase):

    def setUp(self):
        self.storage = storage
        self.storage.reset_mock()
        self.patcher = mock.patch('uwsgi_tasks.tasks.uwsgi')

    def test_runtime_task_execution(self):

        def runtime_task(a, b, c=None):
            self.storage(a, b, c=c)

        runtime_task = RuntimeTask(runtime_task)

        runtime_task(1, 2, c=3)
        self.storage.assert_called_once_with(1, 2, c=3)

        runtime_task(4, '2')
        self.storage.assert_called_with(4, '2', c=None)

        with self.assertRaises(TypeError):
            runtime_task()

    def test_task_is_executed_at_runtime_if_uwsgi_not_available(self):
        mule_task(4, 5, '6', '7')
        self.storage.assert_called_once_with(4, 5, '6', '7')

    def test_mule_task_execution_became_runtime(self):
        # 'mule' not in uwsgi.opt, so it goes to runtime
        with self.patcher:
            mule_task(4, 7, '8')

        self.storage.assert_called_with(4, 7, '8', '4')

    def test_mule_task_execution(self):
        with self.patcher as uwsgi_mock:
            uwsgi_mock.opt = {'mule': 1}
            m_task = mule_task(6, 6,  u'a', u'кириллический')

            message = m_task.get_message_content()
            uwsgi_mock.mule_msg.assert_called_with(
                message, m_task.default_mule_id
            )
            manage_mule_request(message)

        self.storage.assert_called_once_with(6, 6, u'a', u'кириллический')

    def test_spooler_task_execution(self):
        with self.patcher as uwsgi_mock:
            uwsgi_mock.opt = {'spooler': '/tmp/spooler'}
            uwsgi_mock.SPOOL_OK = SPOOL_OK

            s_task = spooler_task(0, '1')
            message = s_task.get_message_content()

            self.assertEqual(message[b'priority'], b'10')
            uwsgi_mock.spool.assert_called_with(message)
            manage_spool_request(message)

        self.storage.assert_called_once_with(0, '1')

    def test_rb_timer_task_execution(self):
        with self.patcher as uwsgi_mock:
            t_task = TimerTask(
                timer_task,
                signal=10,
                seconds=5,
                iterations=3,
                target='workers'
            )

            t_task()

            uwsgi_mock.register_signal.assert_called_once_with(
                10, 'workers', t_task.signal_handler
            )
            uwsgi_mock.add_rb_timer.assert_called_once_with(10, 5, 3)

            t_task.signal_handler(10)
            self.storage.assert_called_once_with(10)

    def test_spooler_retry_on_exception(self):
        with self.patcher as uwsgi_mock:
            uwsgi_mock.opt = {'spooler': '/tmp/spooler'}
            uwsgi_mock.SPOOL_OK = SPOOL_OK

            s_task = spooler_retry_task(666, 777)
            message = s_task.get_message_content()

            self.assertFalse(message.get(b'at'))

            uwsgi_mock.spool.assert_called_with(message)

            manage_spool_request(message)
            self.storage.assert_called_once_with(666, 777)
            self.assertEqual(2, uwsgi_mock.spool.call_count)

            new_message = uwsgi_mock.spool.call_args_list[-1][0][0]
            self.assertTrue(new_message.get(b'at'))

            manage_spool_request(message)

        self.assertEqual(2, self.storage.call_count)

    def test_spooler_task_introspection(self):
        with self.patcher as uwsgi_mock:
            uwsgi_mock.opt = {'spooler': '/tmp/spooler'}

            s_task = spooler_and_task_introspection('a', 1)
            message = s_task.get_message_content()
            manage_spool_request(message)

            new_message = uwsgi_mock.spool.call_args_list[-1][0][0]
            manage_spool_request(new_message)

        self.storage.assert_called_with('a', 1)
        self.assertEqual(2, self.storage.call_count)

    def test_task_working_directory_changed(self):
        current_dir = os.getcwd()

        with self.patcher as uwsgi_mock:
            uwsgi_mock.opt = {'spooler': SPOOLER_DIR}

            s_task = spooler_task_with_cwd_changed()
            message = s_task.get_message_content()
            os.chdir(SPOOLER_DIR)
            manage_spool_request(message)
            self.assertTrue(self.storage.called)
            self.storage.reset_mock()

            os.chdir(current_dir)

            s_task = spooler_task_with_cwd_not_changed()
            message = s_task.get_message_content()
            os.chdir(SPOOLER_DIR)
            manage_spool_request(message)
            self.assertTrue(self.storage.called)
            self.storage.reset_mock()

        os.chdir(current_dir)



# -*- coding: utf-8 -*-
from uwsgi_tasks.tasks import Task, TaskExecutor, TimerTask, CronTask

__all__ = ('task', 'timer', 'timer_lazy', 'cron', 'cron_lazy')


def task(func=None, executor=TaskExecutor.AUTO, **setup):

    def create_task(function):
        return Task(function, executor, **setup)

    if callable(func):
        return create_task(func)
    else:
        return lambda f: create_task(f)


def timer(func=None, seconds=0, iterations=None,
          executor=TaskExecutor.AUTO, **setup):
    """Create timer on initialization"""
    return timer_lazy(
        func=func,
        seconds=seconds,
        iterations=iterations,
        executor=executor,
        run=True,
        **setup
    )


def timer_lazy(func=None, seconds=0, iterations=None,
               executor=TaskExecutor.AUTO, run=False, **setup):
    """Create task on execution"""

    def inner(function):
        timer_task = TimerTask(
            function=function,
            executor=executor,
            seconds=seconds,
            iterations=iterations,
            **setup
        )

        if run:
            return timer_task()
        else:
            timer_task.register_signal()
            return timer_task

    if func is None:
        return inner

    return inner(func)


def cron(func=None, minute=-1, hour=-1, day=-1, month=-1, dayweek=-1,
         executor=TaskExecutor.AUTO, **setup):
    """Creates cron-like task on initialization"""
    return cron_lazy(
        func=func,
        minute=minute,
        hour=hour,
        day=day,
        month=month,
        dayweek=dayweek,
        executor=executor,
        run=True,
        **setup
    )


def cron_lazy(func=None, minute=-1, hour=-1, day=-1, month=-1, dayweek=-1,
              executor=TaskExecutor.AUTO, run=False, **setup):
    """Creates cron-like task on execution"""

    def inner(function):
        cron_task = CronTask(
            function=function,
            executor=executor,
            minute=minute,
            hour=hour,
            day=day,
            month=month,
            dayweek=dayweek,
            **setup
        )

        if run:
            return cron_task()
        else:
            return cron_task

    if func is None:
        return inner

    return inner(func)
